package graph;

import java.util.ArrayList;

import org.jgrapht.DirectedGraph;
import org.jgrapht.graph.DefaultDirectedGraph;

/**
 * The hierarchical structure of the compound graph. This uses the jgrapht library.
 * 
 * @author DarkMatt
 *
 */
public class HierarchyGraph {

	/**
	 * Number of edges.
	 */
	private static Integer edgeCount = 0;
	
	/**
	 * One node of the hierarchy graph.
	 * 
	 * @author DarkMatt
	 *
	 */
	public class HNode {
		public String name = "";
		public float[] position = new float[]{0.0f, 0.0f};
		public int numLeaves;
		
		public HNode(String name) {
			this.name = name;
		}
	}
	
	/**
	 * Root node.
	 */
	private HNode root;
	private DirectedGraph<HNode, Integer> g;
	
	private ArrayList<ConnectionEdge> connections;
	
	public HierarchyGraph() {
		g = new DefaultDirectedGraph<HNode, Integer>(Integer.class);
		
		connections = new ArrayList<ConnectionEdge>();
		
		//g.addVertex("v1");
		//g.addVertex("v2");
		
		//g.addEdge("v1", "v2", "e1");
		
	}
	
	/**
	 * Add the root node.
	 * 
	 * @param root
	 * @return
	 */
	public HNode addRoot(String root) {
		
		this.root = new HNode(root);
		
		// remove file extension
		if (root.indexOf(".") > 0) root = root.substring(0, root.lastIndexOf("."));
		
		g.addVertex(this.root);
		
		return this.root;
	}
	
	/**
	 * Add a node.
	 * 
	 * @param parent
	 * @param node
	 * @return
	 */
	public HNode addNode(HNode parent, String node) {
		
		// remove file extension
		if (node.indexOf(".") > 0) node = node.substring(0, node.lastIndexOf("."));
		
		HNode out = new HNode(node);
		
		//System.out.println("added edge from " + parent.name + " to " + node);
		
		g.addVertex(out);
		g.addEdge(parent, out, edgeCount); // without name
		edgeCount ++;
		
		return out;
	}
	
	/**
	 * One layout.
	 */
	public void layoutGraph(float[] pos) {
		root.position = pos.clone();
		
		positionNodeReArrange(root, new float[]{0.0f, 1.0f});
	}
	
	/**
	 * The vertical range defines how much space is used for children (in an absolute way).
	 * 
	 * @param parentNode
	 * @param horizontalInterval
	 */
	private void positionNode(HNode parentNode, float[] verticalRange) {
		ArrayList<Integer> outEdges = new ArrayList<Integer>(g.outgoingEdgesOf(parentNode));
		
		float rightOffset = 0.15f;
		
		int numChilds = outEdges.size();
		
		float range = verticalRange[1] - verticalRange[0];
		float offs = range/(float)(numChilds);
		
		//global upper position
		float horPosUpper = parentNode.position[1] - range/2.0f;
		for (int i = 0; i < numChilds; i++) {
			
			HNode child = g.getEdgeTarget(outEdges.get(i));
			
			
			float horPos = horPosUpper + offs/2.0f + i*offs;
			child.position = new float[]{parentNode.position[0] + rightOffset, horPos};
			
			//System.out.println(child.name);
			
			ConnectionEdge c = new ConnectionEdge();
			c.path = new float[]{parentNode.position[0], parentNode.position[1], child.position[0], child.position[1]};
			connections.add(c);
			
			positionNode(child, new float[]{horPosUpper + i*offs, horPosUpper + (i+1)*offs});
			
		}
	}
	
	/**
	 * Recursive.
	 * 
	 * The vertical range defines how much space is used for children (in an absolute way).
	 * 
	 * @param parentNode
	 * @param horizontalInterval
	 */
	private void positionNodeReArrange(HNode parentNode, float[] verticalRange) {
		ArrayList<Integer> outEdges = new ArrayList<Integer>(g.outgoingEdgesOf(parentNode));
		
		float rightOffset = 0.15f;
		
		int numChilds = outEdges.size();
		
		float range = verticalRange[1] - verticalRange[0];
		float offs = range/(float)(parentNode.numLeaves); //intervals inside the parent
		
		float prevLeavesPos = 0;
		// global upper position
		float horPosUpper = parentNode.position[1] - range/2.0f;
		for (int i = 0; i < numChilds; i++) {
			
			HNode child = g.getEdgeTarget(outEdges.get(i));
			
			float horPos = horPosUpper + prevLeavesPos + offs*(float)child.numLeaves/2.0f; // + offs/2.0f
			child.position = new float[]{parentNode.position[0] + rightOffset, horPos}; // every child gets as much space as it has leaves
			
			//System.out.println(child.name);
			
			ConnectionEdge c = new ConnectionEdge();
			c.path = new float[]{parentNode.position[0], parentNode.position[1], child.position[0], child.position[1]};
			connections.add(c);
			
			positionNodeReArrange(child, new float[]{horPosUpper + prevLeavesPos, horPosUpper + prevLeavesPos + offs*(float)child.numLeaves});
			
			prevLeavesPos += offs*(float)child.numLeaves;
			
		}
	}
	
	/**
	 * Gets the number of leaves of a hierarchy graph. Every leave gets the total amount of leaves along its branches.
	 */
	public void reArrangeGraph() {
		// count children
		getNumLeaves(root);
	}
	
	/**
	 * Get number of leaves recursion.
	 * 
	 * @param parentNode
	 * @return
	 */
	private int getNumLeaves(HNode parentNode) {
		
		int numLeaves = 0;
		
		ArrayList<Integer> outEdges = new ArrayList<Integer>(g.outgoingEdgesOf(parentNode));
		for (int i = 0; i < outEdges.size(); i++) {
			HNode child = g.getEdgeTarget(outEdges.get(i));
			numLeaves += getNumLeaves(child);
		}
		
		if (outEdges.size() == 0) {
			parentNode.numLeaves = 1;
			return 1;
		} else {
			parentNode.numLeaves = numLeaves;
			return numLeaves;
		}
	}
	
	/**
	 * Gets the positions of the nodes.
	 * 
	 * @return
	 */
	public float[] getPoints() {
		ArrayList<HNode> nodes = new ArrayList<HNode>(g.vertexSet());
		
		float[] out = new float[2 * nodes.size()];
		
		for (int i = 0; i < nodes.size(); i++) {
			out[2*i] = nodes.get(i).position[0];
			out[2*i+1] = nodes.get(i).position[1];
		}
		
		return out;
	}
	
	/**
	 * Gets the labels of the nodes.
	 * 
	 * @return
	 */
	public String[] getPointLabels() {
		
		ArrayList<HNode> nodes = new ArrayList<HNode>(g.vertexSet());
		
		String[] out = new String[nodes.size()];
		
		for (int i = 0; i < out.length; i++) {
			out[i] = nodes.get(i).name;
		}
		
		return out;
	}
	
	/**
	 * Finds the first occurrence of a name.
	 * 
	 * TODO - problem same named files can be a problem!
	 * 
	 * @param name
	 * @return
	 */
	public HNode findNode(String name) {
		
		ArrayList<HNode> nodes = new ArrayList<HNode>(g.vertexSet());
		
		for (int i = 0; i < nodes.size(); i++) {
			if (nodes.get(i).name.equals(name)) {
				return nodes.get(i);
			}
		}
		
		return null;
	}
	
	/**
	 * Returns the edges.
	 * 
	 * @return
	 */
	public ArrayList<ConnectionEdge> getEdges() {
		
		return connections;
	}
	
}
