package renderer;

import java.nio.FloatBuffer;
import java.nio.ShortBuffer;
import java.util.Arrays;

import org.lwjgl.BufferUtils;
import org.lwjgl.opengl.GL11;

import renderer.shader.ShaderCollection;

/**
 * Draws a cubic uniform b-spline. This can also be used for non-uniform control points 
 * (the result is only slightly incorrect, if the distances don't differ too much).
 * 
 * @author DarkMatt
 *
 */
public class CubicBSpline {

	/**
	 * Control points
	 */
	private float[] controlPoints;
	
	/**
	 * Sample density (smoothness of the curve).
	 */
	private int numSamples;
	/**
	 * This offset determines where in the texture collection the spline textures start.
	 * (Probably not the nicest solution.)
	 */
	private int textureOffset;
	/**
	 * Line thickness of a spline.
	 */
	private float splineThickness = 1.5f;
	
	/**
	 * Loads the control points for a spline. Beta defines an smoothing factor of the control points as described in:
	 * http://www.win.tue.nl/~dholten/papers/bundles_infovis.pdf
	 * 
	 * This method calls the 3D version.
	 * 
	 * @param beta
	 * @param points
	 * @param numSamples
	 * @param textures
	 * @param vbo
	 */
	public void loadControlPoints2D(float beta, float[] points, int numSamples, TextureCollection textures, VBO vbo) {
		
		float[] toAdd = control2Dto3D(points);
		
		this.loadControlPoints3D(beta, toAdd, numSamples, textures, vbo);
		
	}
	
	/**
	 * Loads the control points for a spline.
	 * 
	 * This method calls the 3D version.
	 * 
	 * @param points
	 * @param numSamples
	 * @param textures
	 * @param vbo
	 */
	public void loadControlPoints2D(float[] points, int numSamples, TextureCollection textures, VBO vbo) {
		
		float[] toAdd = control2Dto3D(points);
		
		this.loadControlPoints3D(toAdd, numSamples, textures, vbo);
		
	}
	
	/**
	 * Converts 2D control points to 3D.
	 * 
	 * @param points
	 * @return
	 */
	private float[] control2Dto3D(float[] points) {
		float[] toAdd = new float[(points.length/2)*3];
		
		for (int i = 0; i < toAdd.length; i += 3) {
			toAdd[i] = points[i/3*2];
			toAdd[i+1] = points[(i)/3*2+1];
			toAdd[i+2] = 0.0f;
		}
		return toAdd;
	}
	
	/**
	 * Loads the control points for a spline. Beta defines an smoothing factor of the control points.
	 * 
	 * @param beta
	 * @param points
	 * @param numSamples
	 * @param textures
	 * @param vbo
	 */
	public void loadControlPoints3D(float beta, float[] points, int numSamples, TextureCollection textures, VBO vbo) {
		// also straightens the control points
		this.loadControlPoints3D(straightenSpline(beta, points), numSamples, textures, vbo);
	}
	
	/**
	 * Loads the control points for a spline. Main implementation.
	 * 
	 * @param beta
	 * @param points
	 * @param numSamples
	 * @param textures
	 * @param vbo
	 */
	public void loadControlPoints3D(float[] points, int numSamples, TextureCollection textures, VBO vbo) {
		
		this.numSamples = numSamples;
		this.loadVBO(vbo);
		
		// add caps (2 each)
		float[] first = Arrays.copyOf(points, 3);
		float[] last = Arrays.copyOfRange(points, points.length-3, points.length);
		
		controlPoints = new float[3*2 + points.length + 3*2];
		
		for (int i = 0; i < 6; i++) {
			controlPoints[i] = first[i%3];
		}
		for (int i = 0; i < points.length; i++) {
			controlPoints[i+6] = points[i];
		}
		for (int i = 0; i < 6; i++) {
			controlPoints[points.length+6+i] = last[i%3];
		}
		
		//System.out.println("CubicBSpline: loaded control points");
		
		subDivide3D(controlPoints, textures);
		
	}
	
	/**
	 * Generates a VBO for the spline. This only contains vertices that will be positioned by the shader.
	 * Unnecessary to initialize for every Spline.
	 * 
	 * @param vbo
	 */
	private void loadVBO(VBO vbo) {
		
		FloatBuffer vertices = BufferUtils.createFloatBuffer(3*numSamples);
		float[] verts = new float[3*numSamples];
		for (int i = 0; i < verts.length; i+=3) {
			verts[i] = (float)i/(float)(verts.length-3);
			verts[i+1] = 0.5f;
			verts[i+2] = 0.0f;
		}
		
		vertices.put(verts);
		vertices.flip(); // reading mode, position 0
		
		FloatBuffer colors = BufferUtils.createFloatBuffer(3*numSamples);
		float[] cols = new float[3*numSamples];
		for (int i = 0; i < cols.length; i+=3) {
			cols[i] = 1.0f;
			cols[i+1] = 0.0f;
			cols[i+2] = 0.0f;
		}
		colors.put(cols);
		colors.flip(); // reading mode, position 0
		
		ShortBuffer indices = BufferUtils.createShortBuffer(numSamples);
		short[] inds = new short[numSamples];
		for (int i = 0; i < inds.length; i++) {
			inds[i] = (short)i;
		}
		indices.put(inds);
		indices.flip(); // reading mode, position 0
		
		vbo.init(0, vertices, colors, indices);
		
	}
	
	/**
	 * Generates sub arrays of textures for control points.
	 * 
	 * @param points
	 * @param textures
	 */
	private void subDivide3D(float[] points, TextureCollection textures) {
		textureOffset = textures.getSize();
		
		for (int i = 0; i < (points.length/3)-3; i++) {
			float[] subArray = Arrays.copyOfRange(points, 3*i, 3*i+3*4);
			textures.addTextureQueue(subArray, subArray.length/3, 1);
		}
	}
	
	/**
	 * Straightens the spline according to the paper.
	 * 
	 * @param beta
	 * @param points
	 * @return
	 */
	private float[] straightenSpline(float beta, float[] points) {
		int li = points.length-3;
		float[] out = points.clone();
		for (int i = 0; i < points.length; i+=3) {
			out[i  ] = beta*points[i  ] + (1-beta) * (points[0] + (float)(i/3)/(float)(points.length/3-1) * (points[li  ]-points[0]));
			out[i+1] = beta*points[i+1] + (1-beta) * (points[1] + (float)(i/3)/(float)(points.length/3-1) * (points[li+1]-points[1]));
			out[i+2] = beta*points[i+2] + (1-beta) * (points[2] + (float)(i/3)/(float)(points.length/3-1) * (points[li+2]-points[2]));
		}
		return out;
	}
	
	/**
	 * Draws the spline.
	 * 
	 * @param viewMatrix
	 * @param textures
	 * @param shaders
	 * @param vbo
	 * @param colorOverride
	 */
	public void drawSpline(float[] viewMatrix, TextureCollection textures, ShaderCollection shaders, VBO vbo, float colorOverride) {
		
		// debug
		/*for (int i = 0; i < (controlPoints.length/3)-1; i++) {
			GL11.glColor3f(0.0f, 1.0f, 0.0f);
			GL11.glBegin(GL11.GL_LINES);
			GL11.glVertex2f(controlPoints[3*i], controlPoints[3*i+1]);
			GL11.glVertex2f(controlPoints[3*(i+1)], controlPoints[3*(i+1)+1]);
			GL11.glEnd();
		}*/
		
		float[] modelMatrix = viewMatrix.clone();
		
		float alpha = 0.3f * (float)Math.pow(6.0, 1.5)/(float)Math.pow(((float)controlPoints.length/3.0f), 1.5); //7 => 1, 7 => 4/5
		
		if (colorOverride == 0.0f) {
			alpha *= 0.5f;
			GL11.glLineWidth(splineThickness/2.0f);
		} else {
			alpha *= 2.0f;
			GL11.glLineWidth(splineThickness);
		}
		
		shaders.bind(ShaderCollection.BSPLINE);
		for (int i = 0; i < (controlPoints.length/3)-3; i++) {
			textures.bind_texture(textureOffset + i);
			shaders.setShaderParameters(
					"controlPoints", 0,
					"modelMatrix", modelMatrix,
					"colorOffset", (float)i * 1.0f/((float)controlPoints.length/3-3),
					"colorScale", 1.0f/((float)controlPoints.length/3-3),
					"colorOverride", colorOverride,
					"alpha", alpha
				);
			vbo.render(0, GL11.GL_LINE_STRIP, numSamples, ShaderCollection.BSPLINE, shaders);
		}
		shaders.unbind(ShaderCollection.BSPLINE);
	}
	
	/**
	 * Sets the thickness of a spline.
	 * 
	 * @param thickness
	 */
	public void setSplineThickness(float thickness) {
		splineThickness = thickness;
	}
	
	/**
	 * Gets the number of control points.
	 * 
	 * @return
	 */
	public int getSize() {
		return controlPoints.length/3;
	}
	
}
