package renderer;

import graph.ConnectionEdge;

import interaction.CollisionDetection;

import java.awt.Canvas;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.util.ArrayList;

import org.lwjgl.LWJGLException;
import org.lwjgl.input.Mouse;
import org.lwjgl.opengl.Display;
import org.lwjgl.opengl.DisplayMode;
import org.lwjgl.opengl.GL11;
import org.lwjgl.util.vector.Vector2f;
import org.lwjgl.util.vector.Vector3f;

import com.infovis.callgraph.CallBackInterface;

import renderer.shader.ShaderCollection;
import tools.Matrix;

/**
 * The main renderer.
 * 
 * @author DarkMatt
 *
 */
public class RenderThread extends Thread {
	
	private final int CLICKTIMER = 250; // this can make problems at los framerates
	
	private boolean running = true;
	private Canvas canvas;
	private boolean resized = false;
	private int sizeX, sizeY;
	
	private TextureCollection textures;
	private TextRenderer textRender;
	
	private boolean graphInitialized;
	private boolean splineInitialized;
	private float[] points;
	private String[] labels;
	private ArrayList<ConnectionEdge> connections;
	private ArrayList<ConnectionEdge> hConnections;
	
	private ShaderCollection shaders;
	private VBO vbo;
	
	private SplineCollectionSelectable splines;
	
	private int selection = -1;
	private CollisionDetection collDet;
	private Matrix viewMatrix = new Matrix();
	private float textSize = 0.12f;
	
	private boolean getTime;
	private long time;
	
	private CallBackInterface cbi;
	
	/**
	 * Constructor. Sets up non OpenGL related initializations. The init() method needs to be called through the render thread.
	 * 
	 * @param canvas
	 * @param sizeX
	 * @param sizeY
	 * @param cbi
	 */
	public RenderThread(Canvas canvas, int sizeX, int sizeY, CallBackInterface cbi) {
		this.canvas = canvas;
		graphInitialized = false;
		splineInitialized = false;
		
		// listener for getting resize events
		canvas.addComponentListener(new ComponentListener() {
			public void componentHidden(ComponentEvent arg0) {}
			public void componentMoved(ComponentEvent arg0) {}
			public void componentShown(ComponentEvent arg0) {}
			@Override
			public void componentResized(ComponentEvent arg0) {
				resized = true;
			}
		});
		
		this.sizeX = sizeX;//canvas.getParent().getWidth(); is 0!
		this.sizeY = sizeY;//canvas.getParent().getHeight();
		
		points = new float[0];
		labels = new String[0];
		connections = new ArrayList<ConnectionEdge>();
		hConnections = new ArrayList<ConnectionEdge>();
		
		this.cbi = cbi;
	}
	
	/**
	 * Initialize OpenGL. This is executed in the render thread.
	 */
	private void init() {
		
		try {
			Display.setFullscreen(false);
			Display.setDisplayMode(new DisplayMode(sizeX, sizeY));
			Display.setParent(canvas);
		} catch (LWJGLException e) {
			e.printStackTrace();
		}
		
		try {
			Display.create();
		} catch (LWJGLException e) {
			e.printStackTrace();
		}
		
		shaders = new ShaderCollection();
		vbo = new VBO();
		
		textures = new TextureCollection();
		
		textRender = new TextRenderer("icons/font.png", textures, 0); // 0 offset (empty textures when text is initialized)
		
		//loadSpline();
		
		textures.addTextureQueue("test");
		
		//textures.uploadTextures();
		
		GL11.glViewport(0, 0, sizeX, sizeY);
		
		//System.out.println("screen dim " + sizeX + "/" + sizeY);
		
		viewMatrix = new Matrix();
		viewMatrix.translate(new Vector2f(-0.9f, 1.05f));
		viewMatrix.scale(new Vector3f(1.5f, -2.0f, 1.0f));
		
		GL11.glHint(GL11.GL_LINE_SMOOTH_HINT, GL11.GL_NICEST);
        GL11.glEnable(GL11.GL_LINE_SMOOTH);
        //GL11.glLineWidth(1.25f);
        GL11.glLineWidth(1.75f);
	}
	
	/**
	 * Generates the splines.
	 */
	private void loadSpline() {
		
		// resets textures..
		textures = new TextureCollection();
		textRender = new TextRenderer("icons/font.png", textures, 0); // 0 offset (empty textures when text is initialized)
		textures.addTextureQueue("test");
		// end
		
		splines = new SplineCollectionSelectable();
		for (int i = 0; i < connections.size(); i++) {
			float[] line = connections.get(i).path;
			
			splines.add2D(connections.get(i).nodes, line, textures, vbo);
			
		}
		
		System.out.println("RenderThread: number of splines " + connections.size());
		
	}
	
	/**
	 * Sets the graph data.
	 * 
	 * @param points
	 * @param labels
	 * @param connections
	 * @param hConnections
	 */
	public void setGraph(float[] points, String[] labels, ArrayList<ConnectionEdge> connections, ArrayList<ConnectionEdge> hConnections) {
		graphInitialized = true;
		splineInitialized = false;
		
		selection = -1;
		collDet = new CollisionDetection();
		time = System.currentTimeMillis();
		
		//System.out.println("RenderThread: got data!");
		this.points = points;
		this.labels = labels;
		this.connections = connections;
		this.hConnections = hConnections;
	}
	
	/**
	 * Render loop.
	 */
	public void run() {
		
		init();
		
		while (running) {
			
			if (resized) {
				sizeX = canvas.getSize().width;
				sizeY = canvas.getSize().height;
				GL11.glViewport(0, 0, sizeX, sizeY);
				resized = false;
			}
			
			if (graphInitialized) {
				if (!splineInitialized) {
					loadSpline();
					textures.uploadTextures();
					splineInitialized = true;
				}
			}
			
			float[] currentMatrix = viewMatrix.getMatrix();	
			/*// mirror y
			currentMatrix[5] *= -1;
			// scale x
			currentMatrix[0] *= 1.5f;
			// scale y
			currentMatrix[5] *= 2.0f;
			// translate x
			currentMatrix[3] -= 1.05f;
			// translate y
			currentMatrix[7] += 1.00f;*/
			
			if (graphInitialized) {
			
				GL11.glDisable(GL11.GL_DEPTH_TEST);
				
				GL11.glMatrixMode(GL11.GL_PROJECTION);
		        GL11.glLoadIdentity();
		        GL11.glOrtho(-1, 1, -1, 1, 1, -1);
		        GL11.glMatrixMode(GL11.GL_MODELVIEW);
		        GL11.glLoadIdentity();
				
		        GL11.glClearColor(1.0f, 0.0f, 0.0f, 1.0f);
				GL11.glClear(GL11.GL_COLOR_BUFFER_BIT | GL11.GL_DEPTH_BUFFER_BIT);
				
				GL11.glColor3f(0.98f, 1.0f, 1.0f);
				GL11.glDisable(GL11.GL_TEXTURE_2D);
				
				GL11.glBegin(GL11.GL_QUADS);
				GL11.glTexCoord2f(1.0f, 0.0f);
				GL11.glVertex2f(1.0f, 1.0f);
				GL11.glTexCoord2f(0.0f, 0.0f);
				GL11.glVertex2f(-1.0f, 1.0f);
				GL11.glTexCoord2f(0.0f, 1.0f); 
				GL11.glVertex2f(-1.0f, -1.0f);
				GL11.glTexCoord2f(1.0f, 1.0f);
				GL11.glVertex2f(1.0f, -1.0f);
				GL11.glEnd();
				
				//System.out.println(selection);
				
				// TODO temporary
				if (selection == -1) {
					int count = splines.drawAll(currentMatrix, textures, shaders, vbo);
					cbi.setActiveEdgeCount(count);
				} else {
					int count = splines.drawAll(labels[selection], currentMatrix, textures, shaders, vbo);
					cbi.setActiveEdgeCount(count);
				}
				
				// draw connections (hierarchy control polygon)
				for (int i = 0; i < hConnections.size(); i++) {
					float[] line = hConnections.get(i).path;
					
					GL11.glDisable(GL11.GL_TEXTURE_2D);
					//GL11.glEnable(GL11.GL_TEXTURE_2D);
					
					for (int j = 0; j < (line.length-2); j += 2) {
						
						GL11.glColor4f(0.0f, 0.0f, 0.0f, 0.15f);
						GL11.glBegin(GL11.GL_LINES);
						
						float[] pos = new float[]{line[j], line[j+1]};
						pos[0] = currentMatrix[0]*pos[0] + currentMatrix[3] + currentMatrix[0] * 0.005f; // Offset!
						pos[1] = currentMatrix[5]*pos[1] + currentMatrix[7];
						float[] pos1 = new float[]{line[j+2], line[j+3]};
						pos1[0] = currentMatrix[0]*pos1[0] + currentMatrix[3];
						pos1[1] = currentMatrix[5]*pos1[1] + currentMatrix[7];
						
						// idea: use vector
						Vector2f vec = new Vector2f(pos1[0] - pos[0], pos1[1] - pos[1]);
						vec.normalise();
						vec.scale(0.05f);
						
						pos[0] = pos[0] + vec.x;
						pos[1] = pos[1] + vec.y;
						
						GL11.glVertex2f(pos[0], pos[1]);
						
						GL11.glVertex2f(pos1[0], pos1[1]);
						
						//System.out.println("line " + line[j]+"/"+line[j+1] + " to " + line[j+2]+"/"+line[j+3]);
						
						GL11.glEnd();
					}
				}
				
				
				GL11.glEnable(GL11.GL_TEXTURE_2D);
				
				collDet.reset();
						
				// draw string (graph)
				GL11.glColor3f(0.0f, 0.0f, 0.0f);
				for (int i = 0; i < labels.length; i++) {
					
					float[] pos = new float[]{points[2*i], points[2*i+1]};
					
					//pos[0] = -1.0f + pos[0]*1.5f*viewMatrix[0];
					//pos[1] = 2.0f*(1.0f-pos[1])-1.0f;
					//pos[1] *= viewMatrix[5];
					
					//pos[0] += viewMatrix[3];
					//pos[1] += viewMatrix[7];
					
					pos[0] = currentMatrix[0]*pos[0] + currentMatrix[3];
					pos[1] = currentMatrix[5]*pos[1] + currentMatrix[7];
					
					collDet.addElement(new float[]{pos[0] + 0.005f, pos[1] - 0.013f, 0.09f, (float)sizeX/(float)sizeY*0.022f});
					
					if (i == selection) {
						GL11.glColor3f(0.0f, 0.7f, 0.7f);
					}
					
					textRender.draw_string(labels[i], pos[0] + 0.005f, pos[1] - 0.015f, textSize, (float)sizeY/(float)sizeX); //.25f
					
					if (i == selection) {
						GL11.glColor3f(0.0f, 0.0f, 0.0f);
					}
					
					//textRender.draw_string_centered(labels[i], -1.0f + pos[0]*1.5f, 2.0f*(1.0f-pos[1])-1.0f, 0.25f, (float)sizeY/(float)sizeX);
				}
				
				/*
				// draw connections (control polygon)
				for (int i = 0; i < connections.size(); i++) {
					float[] line = connections.get(i).path;
					
					GL11.glDisable(GL11.GL_TEXTURE_2D);
					//GL11.glEnable(GL11.GL_TEXTURE_2D);
					
					for (int j = 0; j < (line.length-2); j += 2) {
						
						float offs = (float)i/50.0f;
						
						GL11.glLineWidth(2.5f); 
						GL11.glColor3f(1.0f-(float)(i)/(float)connections.size(), 0.0f, 0.0f);
						GL11.glBegin(GL11.GL_LINES);
						GL11.glVertex2f(1.5f*line[j]-1.0f, 		2.0f*(1.0f-line[j+1])-1.0f + offs);
						GL11.glVertex2f(1.5f*line[j+2]-1.0f, 	2.0f*(1.0f-line[j+3])-1.0f + offs);
						
						//System.out.println("line " + line[j]+"/"+line[j+1] + " to " + line[j+2]+"/"+line[j+3]);
						
						GL11.glEnd();
					}
				}
				*/
				
				// debug
				//collDet.drawBB();
				
			}
			
			/*while (Keyboard.next()) {
				if (Keyboard.getEventKeyState()) {
					if (Keyboard.getEventKey() == Keyboard.KEY_A) {
						System.out.println("A Key Pressed");
					}
					if (Keyboard.getEventKey() == Keyboard.KEY_S) {
						System.out.println("S Key Pressed");
					}
					if (Keyboard.getEventKey() == Keyboard.KEY_D) {
						System.out.println("D Key Pressed");
					}
					} else {
					if (Keyboard.getEventKey() == Keyboard.KEY_A) {
						System.out.println("A Key Released");
					}
					if (Keyboard.getEventKey() == Keyboard.KEY_S) {
						System.out.println("S Key Released");
					}
					if (Keyboard.getEventKey() == Keyboard.KEY_D) {
						System.out.println("D Key Released");
					}
				}
			}*/
			
			if (Mouse.isButtonDown(0)) {
				
				//System.out.println("Mouse pressed");
				
				if (!getTime) {
					time = System.currentTimeMillis();
					getTime = true;
				}
				
				// panning mode
					
				float dx = (float)Mouse.getDX();
				float dy = (float)Mouse.getDY();
					
				viewMatrix.translate(new Vector2f(dx/700.0f, -dy/500.0f*-1.0f/viewMatrix.m11));
				
			} else {
				if (getTime && (System.currentTimeMillis() - time) < CLICKTIMER) {
					// clicking mode only of click was short
					
					float x = (float)Mouse.getX()/(float)sizeX; // normalized
					float y = (float)Mouse.getY()/(float)sizeY;
					x = 2.0f*x -1.0f;
					y = 2.0f*y -1.0f;
					
					int newSelection = collDet.getCollisionIndex(new float[]{x, y}, new float[]{1.0f, 1.0f});
					if (newSelection != selection) selection = newSelection;
					else selection = -1;
					
					//System.out.println("Selection at " + x + "/" + y);
				}
				
				getTime = false;
			}
			
			
			// zoom
			int dWheel = Mouse.getDWheel();
			if (dWheel != 0) {
				
				//System.out.println("MOUSE WHEEL @ X: " + x + " Y: " + y);
				
				if (dWheel < 0) {
					//Vector2f t = new Vector2f(viewMatrix.m30, -0.5f + (1.05f-viewMatrix.m31));
					
					/*Vector4f mp = new Vector4f(0.5f, -0.5f, 0.0f, 1.0f);
					Matrix4f tv = new Matrix4f(viewMatrix);
					Matrix.transform((Matrix4f)tv.transpose(), mp, mp);
					// the draw area is between 0 and 1*/
					//Vector2f t = new Vector2f(avgPos[0], avgPos[1]);
					
					//System.out.println("wheel along " + t.y + " y " + y + " and " + sizeY + " viewmat " + viewMatrix.m31);
					
					//viewMatrix = (Matrix)viewMatrix.translate((Vector2f)t.negate());
					viewMatrix = (Matrix)viewMatrix.translate(new Vector2f(0.0f, -0.5f));
					viewMatrix = (Matrix)viewMatrix.scale(new Vector3f(1.0f, 1.1f, 1.0f));
					viewMatrix = (Matrix)viewMatrix.translate(new Vector2f(0.0f, 0.5f));
					viewMatrix = (Matrix)viewMatrix.translate(new Vector2f(0.0f, 1.0f/1.1f-1.0f));
					//viewMatrix[0] *= 1.1f;
					/*float tmp = viewMatrix[7];
					float mpy = (float)y + tmp;
					viewMatrix[5] *= 1.1f;*/
					//viewMatrix[10] *= 1.1f;
				} else if (dWheel > 0) {
					//Vector2f t = new Vector2f(viewMatrix.m30, -0.5f + (1.05f-viewMatrix.m31));
					/*Vector4f mp = new Vector4f(0.5f, -0.5f, 0.0f, 1.0f);
					Matrix4f tv = new Matrix4f(viewMatrix);
					Matrix.transform((Matrix4f)tv.transpose(), mp, mp);
					// the draw area is between 0 and 1*/
					//Vector2f t = new Vector2f(avgPos[0], avgPos[1]);
					
					//System.out.println("wheel along " + t.y + " y " + y + " and " + sizeY + " viewmat " + viewMatrix.m31);
					
					//viewMatrix = (Matrix)viewMatrix.translate((Vector2f)t.negate());
					viewMatrix = (Matrix)viewMatrix.translate(new Vector2f(0.0f, -0.5f));
					viewMatrix = (Matrix)viewMatrix.scale(new Vector3f(1.0f, 0.9f, 1.0f));
					viewMatrix = (Matrix)viewMatrix.translate(new Vector2f(0.0f, 0.5f));
					viewMatrix = (Matrix)viewMatrix.translate(new Vector2f(0.0f, 1.0f/0.9f-1.0f));
					//viewMatrix = (Matrix)viewMatrix.translate((Vector2f)t.negate().scale(0.9f));
					//viewMatrix[0] *= 0.9f;
					/*viewMatrix[5] *= 0.9f;*/
					//viewMatrix[10] *= 0.9f;
				}
			}
			
			//GL11.glColor3f(0.0f, 0.0f, 0.0f);
			//textRender.draw_string_centered("Some test", 0.0f, 0.0f, 2.0f);
			
			Display.update();
		
		}
		
		Display.destroy();
		
	}
	
	/**
	 * Shut down renderer method.
	 */
	public void terminate() {
		running = false;
	}
	
	/**
	 * Sets the size of the font.
	 * 
	 * @param size
	 */
	public void setTextSize(float size) {
		textSize = size;
	}
	
	/**
	 * Sets the thickness of splines.
	 * 
	 * @param thickness
	 */
	public void setSplineThickness(float thickness) {
		splines.setSplineThickness(thickness);
	}
	
	/**
	 * Activates drawing of focus + context splines (grey splines for context when a selection is taken).
	 * 
	 * @param context
	 */
	public void setSplineContext(boolean context) {
		splines.setEnableContext(context);
	}
	
}
