package renderer;

import java.awt.Color;
import java.awt.Font;
import java.awt.FontFormatException;
import java.awt.FontMetrics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.net.URL;

import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Platform;
import org.lwjgl.opengl.GL11;
import org.osgi.framework.Bundle;

/**
 * 
 * Text renderer.
 * 
 * http://en.wikibooks.org/wiki/OpenGL_Programming/Modern_OpenGL_Tutorial_Text_Rendering_02
 * 
 * 
 * To display strings on the screen.
 * '/r' ... sequences allow custom coloring
 * 
 * @author DarkMatt
 *
 */

public class TextRenderer {
	
	private static final int MAX_HEIGHT = 128;
	private static final int MAX_LENGTH = 4096;
	private static final String CHAROFFSET = "  ";
	
	private class StringInt {
		public String string;
		public int measureSize;
	}
	
	private TextureCollection textures;
	
	/*private static final int FIRST_ROW_OFFSET = 6;
	private static final int ROW_OFFSET = 2;
	private static final int ROWS = 6;*/
	
	private String[] c = new String[]{
			"A","B","C","D","E","F","G","H","I","J","K","L","M","N", 	//14
			"O","P","Q","R","S","T","U","V","W","X","Y","Z",			//12
			"a","b","c","d","e","f","g","h","i","j","k","l","m","n",	//14
			"o","p","q","r","s","t","u","v","w","x","y","z",			//12
			"1","2","3","4","5","6","7","8","9","0",					//10
			"*","+","'","#",",",";",".",":","_","-"};					//10
	private int[] offsetInt;
	
	//private int[] offset; //sum - 6
	
	private float[] cx1, cy1, cx2, cy2, cxnext;
	
	private int textureIndex;
	
	/**
	 * Constructor. Needs the texture collection and the index with the last written texture.
	 * 
	 * @param path
	 * @param text
	 * @param textureIndex
	 */
	public TextRenderer(String path, TextureCollection text, int textureIndex) {
		
		textures = text;
		//could be stored in a text file!
		this.textureIndex = textureIndex;
		
		offsetInt = new int[2*c.length+2];
		
		generateFont(); // TODO not necessary (possible) in a jar file
		
		text.addTextureQueue("customFont");
		
	}
	
	/**
	 * Loads a font from a font file (e.g. *.tff).
	 * 
	 * @param g
	 * @return
	 */
	public StringInt loadFont(Graphics2D g) {
		
		//File f = resolveAbsolutePath("icons/Lobster/Lobster_1.3.otf");
		URL url = resolveAbsolutePath("icons/Droid-Sans/DroidSans.ttf");
		//File f = resolveAbsolutePath("icons/source-sans-pro/SourceSansPro-Regular.otf");
		//File f = resolveAbsolutePath("icons/architects-daughter/ArchitectsDaughter.ttf");
		
		//cbi.setActiveEdgeCount(333);
		
		Font font = null;
		if (url != null) {
			try {
				font = Font.createFont(Font.TRUETYPE_FONT, url.openStream());
			} catch (FontFormatException | IOException e1) {
				throw new NullPointerException("No url " + url.getPath());
			}
		}
		font = font.deriveFont(Font.PLAIN, 22);
		
		g.setFont(font);
		
		FontMetrics metrics = g.getFontMetrics(font);
		
		// iterate over the string to get the offsets for letters
		// A B C .. always followed by a space character to avoid overlapping letters
		int currentLength = 0;
		offsetInt[0] = 0;
		String currentString = c[0];
		offsetInt[1] = metrics.stringWidth(c[0]);
		for (int i = 1; i < c.length; i++) {
			currentString += CHAROFFSET;
			currentLength = metrics.stringWidth(currentString);
			offsetInt[2*i] = currentLength;
			currentString += c[i];
			currentLength = metrics.stringWidth(currentString);
			offsetInt[2*i+1] = currentLength;
		}
		currentString += CHAROFFSET;
		currentLength = metrics.stringWidth(currentString);
		offsetInt[offsetInt.length-2] = currentLength;
		currentLength = metrics.stringWidth(currentString);
		offsetInt[offsetInt.length-1] = currentLength;
		
		// in the offsetInt array the positions of:
		// 1) character start at %2 == 0
		// 2) next character position at %2 == 1
		
		// save data to arrays
		cx1 = new float[c.length];
		cy1 = new float[c.length];
		cx2 = new float[c.length];
		cy2 = new float[c.length];
		cxnext = new float[c.length];
		
		float nextPowerTwo = (float)Math.pow(2, Math.ceil(Math.log(currentLength)/Math.log(2)));
		for (int i = 0; i < cx1.length; i++) {
			cx1[i] = (float)offsetInt[2*i]/nextPowerTwo;
			cx2[i] = (float)offsetInt[2*i+2]/(float)nextPowerTwo;
			cxnext[i] = (float)offsetInt[2*i+1]/nextPowerTwo;
			cy1[i] = 0.0f;
			cy2[i] = 0.20f;
		}
		
		StringInt out = new StringInt();
		out.string = currentString;
		out.measureSize = currentLength;
		return out;
	}
	
	/**
	 * Writes a font to an image and loads it. Writing is disabled in this case.
	 */
	public void generateFont() {
		
		//TODO temporary disabled
		
		/*File imageFile = resolveAbsolutePath("icons/customFont.png");
		try {
			imageFile.createNewFile();
		} catch (IOException e1) {
			e1.printStackTrace();
		}
		
		System.out.println(imageFile.getAbsolutePath());*/
		
		BufferedImage image = new BufferedImage(MAX_LENGTH, MAX_HEIGHT, BufferedImage.TYPE_INT_ARGB);
		
		Graphics2D g = image.createGraphics();
		// anti aliasing
		RenderingHints rh = new RenderingHints(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_GASP);
	    g.setRenderingHints(rh);
	    
		g.setColor(Color.white);
		
		StringInt str = loadFont(g);
		
	    g.drawString(str.string, 0, 20);
		g.dispose();
		
		/*// power of two
		int nextPowerTwo = (int)Math.pow(2, Math.ceil(Math.log(str.measureSize)/Math.log(2)));
		try {
			ImageIO.write(image.getSubimage(0, 0, nextPowerTwo, MAX_HEIGHT), "png", imageFile);
		} catch (IOException e) {
			e.printStackTrace();
		}*/
		
	}
	
	/**
	 * Gets the path of a resource.
	 * 
	 * @param path
	 * @return
	 */
	private URL resolveAbsolutePath(String path) {
		
		//"The classpath for code in a jar is the jarfile which is NOT useable in a File constructor." !!
		
		/*URI uri = null;
		try {
			uri = new URI("file://"+TextRenderer.class.getProtectionDomain().getCodeSource().getLocation().getPath());
		} catch (URISyntaxException e1) {
			throw new RuntimeException("URI fault");
		}
        File f = new File(uri);
        return new File(f.getParent() + "/" + path);*/
		
		Bundle bundle = Platform.getBundle("callgraph");
		Path realPath = new Path(path);
		URL fileURL = FileLocator.find(bundle, realPath, null);
		URL resolved = null;
		try {
		    resolved = FileLocator.resolve(fileURL);
		} catch (IOException e) {
		    throw new RuntimeException(e);
		}
		
		return resolved;
        
	}
	
	/**
	 * Draws a string centered.
	 * 
	 * @param text
	 * @param px
	 * @param py
	 * @param scale
	 * @param aspect
	 */
	public void draw_string_centered(String text, float px, float py, float scale, float aspect) {
		
		float size = 0.0f;
		
		for (int i=0; i<text.length(); i++) {	
			for (int j=0; j<c.length; j++) {
				if (c[j].equals(text.substring(i, i+1))) {
					size += (0.01f + cx2[j] - cx1[j]) * scale;
				} else if (text.substring(i, i+1).equals(" ")){
					size += 0.0003f * scale;
				} else if (text.substring(i, i+1).equals("/")){
					i += 2;
				}
			}
		}
		
		size /= 2.0f;
		
		GL11.glPushMatrix();
		
		GL11.glTranslatef(-size, 0.0f, 0.0f);
		
		draw_string(text, px, py, scale, aspect);
		
		GL11.glPopMatrix();
		
	}
	
	/**
	 * Draws a string.
	 * 
	 * @param text
	 * @param px
	 * @param py
	 * @param scale
	 * @param aspect
	 */
	public void draw_string(String text, float px, float py, float scale, float aspect) {
		
		GL11.glEnable(GL11.GL_TEXTURE_2D);
		GL11.glEnable(GL11.GL_BLEND);
		GL11.glBlendFunc(GL11.GL_SRC_ALPHA, GL11.GL_ONE_MINUS_SRC_ALPHA);
		
		//GL11.glBindTexture(GL11.GL_TEXTURE_2D, font_texture.getTextureID());
		
		//GL11.glEnable(GL11.GL_BLEND);
		//GL11.glBlendFunc(GL11.GL_SRC_ALPHA, GL11.GL_ONE_MINUS_SRC_ALPHA);
		
		textures.bind_texture(textureIndex);
		
		GL11.glPushMatrix();
		
		float dx = 0.0f;
		
		for (int i=0; i<text.length(); i++) {
			
			for (int j=0; j<c.length; j++) {
				
				if (c[j].equals(text.substring(i, i+1))) {
					
					//GL11.glColor3f(1.0f, 1.0f, 1.0f);
					
					//GL11.glColor3f(1.0f, 1.0f, 1.0f); //red try
					
					GL11.glTranslatef(dx + 0.005f*scale, 0.0f, 0.0f);
					
					//System.out.println(cx1[j] +"/"+ cy1[j] +"/"+ cx2[j] +""+ cy2[j]);
					
					//dx = (cx2[j] - cx1[j]);
					dx = 8.0f * scale * (cxnext[j] - cx1[j]);
					float dy = (cy2[j] - cy1[j]);
					
					float dSize = 8.0f * scale * (cx2[j] - cx1[j]);
					dy *= 0.8*scale*(1.0f/aspect);
					
					GL11.glBegin(GL11.GL_QUADS);
					GL11.glTexCoord2f(cx1[j]+0.00005f, cy2[j]);
					GL11.glVertex2f(px, py);
					GL11.glTexCoord2f(cx2[j]-scale*0.003f, cy2[j]);
					GL11.glVertex2f(px+dSize, py);
					GL11.glTexCoord2f(cx2[j]-scale*0.003f, cy1[j]); 
					GL11.glVertex2f(px+dSize, py+dy);
					GL11.glTexCoord2f(cx1[j]+0.00005f, cy1[j]);
					GL11.glVertex2f(px, py+dy);
					/*GL11.glTexCoord2f(cx1[j], cy2[j]);
					GL11.glVertex2f(px, py);
					GL11.glTexCoord2f(cx2[j], cy2[j]);
					GL11.glVertex2f(px+dx, py);
					GL11.glTexCoord2f(cx2[j], cy1[j]); 
					GL11.glVertex2f(px+dx, py+dy);
					GL11.glTexCoord2f(cx1[j], cy1[j]);
					GL11.glVertex2f(px, py+dy);*/
					GL11.glEnd();
					
				} else if (text.substring(i, i+1).equals(" ")){
					GL11.glTranslatef(0.0003f * scale, 0.0f, 0.0f);
				} else if (text.substring(i, i+1).equals("/")){
					i ++;
					if (text.substring(i, i+1).equals("r")){
						GL11.glColor3f(1.0f, 0.0f, 0.0f);
					} else if (text.substring(i, i+1).equals("w")){
						GL11.glColor3f(1.0f, 1.0f, 1.0f);
					} else if (text.substring(i, i+1).equals("g")){
						GL11.glColor3f(0.5f, 0.5f, 0.5f);
					}
					i ++;
				}
				
			}
			
		}
		
		GL11.glPopMatrix();
		
		//GL11.glDisable(GL11.GL_BLEND);
		
	}
	
}
