package renderer;

import java.io.IOException;
import java.nio.FloatBuffer;
import java.util.ArrayList;
import java.util.HashMap;

import org.lwjgl.BufferUtils;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL30;
import org.newdawn.slick.opengl.InternalTextureLoader;
import org.newdawn.slick.opengl.Texture;
import org.newdawn.slick.opengl.TextureLoader;
import org.newdawn.slick.util.Log;
import org.newdawn.slick.util.ResourceLoader;

/**
 * Manages the textures.
 * Used twice:
 *  - general renderer
 *  - loading screen
 *  
 * 
 * Does not support the texture name "..." (probably not a valid file-name anyway)
 * 
 * Implementation:
 *  - HashSet for fast contains
 * 
 * @author DarkMatt
 *
 */

public class TextureCollection {
	
	private class TextureLoadObject {
		public String textFileName;
		public int[] size;
		public float[] data;
		public TextureLoadObject(String textFileName) {
			this.textFileName = textFileName;
		}
		public TextureLoadObject(int[] size, float[] data) {
			this.size = size;
			this.data = data;
		}
	}
	private ArrayList<TextureLoadObject> textLoadData;
	
	/** for the textures read via file */
	private ArrayList<Texture> textures;
	// stores the already read file-names and their positions in the data-structure
	private HashMap<String, Integer> fileNames;

	
	/** maps the indices to either of the two lists */
	private class ArrayIndex {
		public int index;
	}
	private ArrayList<ArrayIndex> indexMapping;
	
	/**
	 * Constructor
	 */
	public TextureCollection() {
		
		Log.setLogSystem(null);
		
		textures = new ArrayList<Texture>();
		fileNames = new HashMap<String, Integer>();
		
		indexMapping = new ArrayList<ArrayIndex>();
		
		// loader object
		textLoadData = new ArrayList<TextureLoadObject>();
		
	}
	
	/**
	 * Adds a texture with a specified path.
	 * 
	 * @param name
	 * @return
	 */
	public int addTextureQueue(String name) {
		
		if (name.length() == 0) {
			return -1;
		}
		
		Integer index = fileNames.get(name); // check if the filename was already load
		if (index != null) {
			return index;
		}
		
		int size = textLoadData.size();
		fileNames.put(name, size);
		textLoadData.add(new TextureLoadObject(name));
		
		return size;
		
	}
	
	/**
	 * Adds a texture from an array.
	 * 
	 * @param data
	 * @param sx
	 * @param sy
	 * @return
	 */
	public int addTextureQueue(float[] data, int sx, int sy) {
		
		int size = textLoadData.size();
		TextureLoadObject tlo = new TextureLoadObject(new int[]{sx, sy}, data);
		textLoadData.add(tlo);
		
		return size;
		
	}
	
	/**
	 * Performs texture upload to the graphics card.
	 */
	public void uploadTextures() {
		
		//GL11.glEnable(GL11.GL_TEXTURE);
		
		for (int i = 0; i < textLoadData.size(); i++) {
			String name = textLoadData.get(i).textFileName;
			if (name == null) {	// add via source code - array
				
				ArrayIndex ai = new ArrayIndex();
				ai.index = textures.size();
				indexMapping.add(ai);
				
				InternalTextureLoader tl = InternalTextureLoader.get();
				
				int size[] = textLoadData.get(i).size;
				
				try {
					textures.add(tl.createTexture(size[0], size[1], GL11.GL_LINEAR));
				} catch (IOException e) {
					e.printStackTrace();
				}
				
				FloatBuffer tex_data = BufferUtils.createFloatBuffer(3*textLoadData.get(i).size[0]*textLoadData.get(i).size[1]); //why 3*2?
				tex_data.put(textLoadData.get(i).data);
				tex_data.position(0);
				
				/*GL11.glEnable(GL11.GL_TEXTURE_2D);
				GL11.glBindTexture(GL11.GL_TEXTURE_2D, texturesFromArray.get(number));*/
				GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MIN_FILTER, GL11.GL_NEAREST);
				GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MAG_FILTER, GL11.GL_NEAREST);
				GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_WRAP_S, GL11.GL_CLAMP);
				GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_WRAP_T, GL11.GL_CLAMP);
				
				textures.get(textures.size()-1).bind();
				GL11.glTexImage2D(GL11.GL_TEXTURE_2D, 0, 
						GL11.GL_RGB, size[0], size[1], 
						0, GL11.GL_RGB, GL11.GL_FLOAT, tex_data);
				
			} else { // add via file - string
				try {
					
					ArrayIndex ai = new ArrayIndex();
					ai.index = textures.size();
					indexMapping.add(ai);
					
					textures.add(
							TextureLoader.getTexture(
									"PNG", 
									ResourceLoader.getResourceAsStream("icons/"+name+".png"),
									GL11.GL_LINEAR)
					);
					
					//EXTFramebufferObject.glGenerateMipmapEXT(GL11.GL_TEXTURE_2D);
					GL30.glGenerateMipmap(GL11.GL_TEXTURE_2D);
					
					//GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MIN_FILTER, GL11.GL_LINEAR);
					GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MIN_FILTER, GL11.GL_LINEAR_MIPMAP_LINEAR);
					GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MAG_FILTER, GL11.GL_LINEAR);
					GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_WRAP_S, GL11.GL_REPEAT);
					GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_WRAP_T, GL11.GL_REPEAT);
					
					//System.out.println("Added texture: " + name + " at " + (debug_counter++));
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
			
		}
		
		//System.out.println("TextureColletion: loading done.");
		
	}
	
	/**
	 * Returns the size of the texture collection.
	 * 
	 * @return
	 */
	public int getSize() {
		return textLoadData.size();
	}
	
	/**
	 * Binds a texture
	 * 
	 * @param index
	 */
	public void bind_texture(int index) {
		textures.get(indexMapping.get(index).index).bind();
	}
	
	/**
	 * Unbinds a texture
	 */
	public void unbind_texture() {
		GL11.glBindTexture(GL11.GL_TEXTURE_2D, 0);
	}
	
}
