package renderer;


import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.nio.ShortBuffer;

import renderer.shader.ShaderCollection;

import org.lwjgl.BufferUtils;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL15;
import org.lwjgl.opengl.GL20;
import org.lwjgl.opengl.Util;

/**
 * 
 * VBO implementation.
 * 
 *  * PROBLEM/COMMENTS:
 * 
 *  impostor lighting
 *  http://www.gamasutra.com/view/feature/130911/dynamic_2d_imposters_a_simple_.php
 * 
 *  Batch-ing
 *  LSP
 *  Light Pre-Pass Renderer
 * 
 *http://diaryofagraphicsprogrammer.blogspot.co.at/
 * 
 * Encapsulates VBO functionality.
 *  - init
 *  - modify/update
 *  - draw
 *  
 *  Currently the VBO maximum size is hard-coded, since adding of objects should be theoretically possible dynamically.
 * 
 * Death is certain
 * 
 * 
 * @author DarkMatt
 *
 */

public class VBO {
	
	protected final int SIZE = 10; //TODO estimated size...
	
	/** keeps track of the number of instances */
	protected int actualSize;
	
	/** Index buffer object */
	private IntBuffer vboIBO;
	/** Data buffer object */
	private IntBuffer vboId;
	
	/**
	 * Constructor
	 */
	public VBO() {
		vboIBO = BufferUtils.createIntBuffer(SIZE);
		vboId = BufferUtils.createIntBuffer(2*SIZE);	// position, color, normal, textCoord, tangent
		
		GL15.glGenBuffers(vboIBO);
		GL15.glGenBuffers(vboId);
		
		actualSize = 0;
	}
	
	/**
	 * Sets the data of the VBO for a single instance/object.
	 * 
	 * @param objectId
	 * @param indices
	 * @param vertices
	 */
	public void update(int objectId, short[] indices, float[] vertices) {
		
		ShortBuffer ind = BufferUtils.createShortBuffer(indices.length);
		ind.put(indices);
		ind.position(0);
		
		GL15.glBindBuffer(GL15.GL_ELEMENT_ARRAY_BUFFER, vboIBO.get(objectId));
		GL15.glBufferData(GL15.GL_ELEMENT_ARRAY_BUFFER, ind, GL15.GL_STATIC_DRAW);
		
		FloatBuffer in = BufferUtils.createFloatBuffer(vertices.length);
		in.put(vertices);
		in.position(0);
		
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, vboId.get(2*objectId+0));
		GL15.glBufferData(GL15.GL_ARRAY_BUFFER, in, GL15.GL_STREAM_DRAW);
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, 0);
		
	}
	
	/**
	 * Changes the vertex positions.
	 * 
	 * @param objectId
	 * @param vertices
	 */
	public void updatePositions(int objectId, float[] vertices) {
		
		
		FloatBuffer in = BufferUtils.createFloatBuffer(vertices.length);
		in.put(vertices);
		in.position(0);
		
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, vboId.get(2*objectId+0));
		GL15.glBufferData(GL15.GL_ARRAY_BUFFER, in, GL15.GL_STREAM_DRAW);
		//GL15.glBufferSubData(GL15.GL_ARRAY_BUFFER, 0, in); //(long)(4*start)
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, 0);
		
	}
	
	/**
	 * Changes the colors (unused).
	 * 
	 * @param objectId
	 * @param color
	 */
	public void updateColors(int objectId, float[] color) {
		
		FloatBuffer in = BufferUtils.createFloatBuffer(color.length);
		in.put(color);
		in.position(0);
		
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, vboId.get(2*objectId+1));
		GL15.glBufferData(GL15.GL_ARRAY_BUFFER, in, GL15.GL_STREAM_DRAW);
		//GL15.glBufferSubData(GL15.GL_ARRAY_BUFFER, 0, in); //(long)(4*start)
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, 0);
		
	}
	
	/**
	 * Delete
	 */
	public void deleteAll() {
		GL15.glDeleteBuffers(vboIBO);
		GL15.glDeleteBuffers(vboId);
	}
	
	/**
	 * Requires the size of the vertex data and the size of the index data!
	 * 
	 * More slick, but outdated version.
	 * 
	 * @param vertices
	 * @param textCoords
	 * @param indices
	 * @param normals
	 * @param vertexCount
	 * @param indicesCount
	 */
	public void init(int objectId, FloatBuffer vertices, FloatBuffer colors, ShortBuffer indices) {
		
		//if (Entrance.DEBUG) System.out.println("5: Initializing VBOs ["+objectId+"]");
		
		//indices
		
		GL15.glBindBuffer(GL15.GL_ELEMENT_ARRAY_BUFFER, vboIBO.get(objectId));
		GL15.glBufferData(GL15.GL_ELEMENT_ARRAY_BUFFER, indices, GL15.GL_STATIC_DRAW);
		
		
		//vertices
		
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, vboId.get(2*objectId+0));
		GL15.glBufferData(GL15.GL_ARRAY_BUFFER, vertices, GL15.GL_STREAM_DRAW);
		
		//colors
		
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, vboId.get(2*objectId+1));
		GL15.glBufferData(GL15.GL_ARRAY_BUFFER, colors, GL15.GL_STREAM_DRAW);
		
		
		
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, 0);
		GL15.glBindBuffer(GL15.GL_ELEMENT_ARRAY_BUFFER, 0);
		
		//if (Entrance.DEBUG) System.out.println("6: Done VBO ["+objectId+"]");
	}
	
	/**
	 * Old basic render call per object
	 * 
	 * @param objectId
	 * @param type
	 * @param indicesSize
	 * @param shaderType
	 * @param sc
	 */
	public void render(int objectId, int type, int indicesSize, int shaderType, ShaderCollection sc) {
		
		int vl = sc.getVertexLocation(shaderType);
		GL20.glEnableVertexAttribArray(vl); // these are not guaranteed to be persistent
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, vboId.get(2*objectId+0));	//vertex pos
		GL20.glVertexAttribPointer(vl, 3, GL11.GL_FLOAT, false, 0, 0);
		
		int cl = sc.getColorLocation(shaderType);
		GL20.glEnableVertexAttribArray(cl);
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, vboId.get(2*objectId+1)); 	//for colors
		GL20.glVertexAttribPointer(cl, 3, GL11.GL_FLOAT, false, 0, 0);
		
		if (indicesSize > 0) {
			
			GL15.glBindBuffer(GL15.GL_ELEMENT_ARRAY_BUFFER, vboIBO.get(objectId)); 	//for indices
			
			try { 
				Util.checkGLError();
			} catch (Exception e) {
				//System.out.println(vl +"/"+ cl +"/"+ tcl +"/"+ nl);
				//System.out.println("pre (" + indicesSize + ") "+ objectId + "/" +shaderType+ " " + e);
			}
			
			//GL40.glPatchParameteri(GL40.GL_PATCH_VERTICES, 3); //TODO only for tessellation
			
			// OLD
			GL11.glDrawElements(type, indicesSize, GL11.GL_UNSIGNED_SHORT, 0);
			
			//GL32.glDrawElementsInstancedBaseVertex(type, indicesSize, GL11.GL_UNSIGNED_SHORT, 0, 10, 0);
			
			try { 
				Util.checkGLError();
			} catch (Exception e) {
				//System.out.println(vl +"/"+ cl +"/"+ tcl +"/"+ nl);
				//System.out.println("post (" + indicesSize + ") "+ objectId + "/" +shaderType+ " " + e);
			}
			
		}
		
		GL20.glDisableVertexAttribArray(vl);
		GL20.glDisableVertexAttribArray(cl);
		
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, 0);
		GL15.glBindBuffer(GL15.GL_ELEMENT_ARRAY_BUFFER, 0);
		
	}
	
}
