package renderer.shader;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.ArrayList;

import org.lwjgl.BufferUtils;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL20;
import org.lwjgl.opengl.GL21;
import org.lwjgl.opengl.GL30;
import org.lwjgl.opengl.GL32;
import org.lwjgl.opengl.GL40;
import org.lwjgl.opengl.GLContext;
import org.lwjgl.opengl.ARBShaderObjects;
import org.lwjgl.opengl.ARBVertexShader;
import org.lwjgl.opengl.ARBFragmentShader;
import org.lwjgl.opengl.Util;

import com.infovis.callgraph.CallBackInterface;

import tools.FileIO;




/**
 * 
 * Implementation for one shader.
 * 
 * This is not OOP - but a mere WRAPPER for Shader OpenGL functionality.
 * 
 * TODO: error checking - this isn't done consistently enough in this project yet.
 * 
 * @author DarkMatt
 *
 */

public class Shader {

	private String name;
	
	private int vertex_shader = 0;
	private int fragment_shader = 0;
	private int geometry_shader = 0;
	private int tess_control_shader = 0;
	private int tess_evaluation_shader = 0;
	private int program = 0;
	// temporary buffer that is used for writing to matrices
	private FloatBuffer matrix = BufferUtils.createFloatBuffer(16);
	
	private boolean debug = false;
	
	//is filled with the UNIFORM location connections
	private ArrayList<Integer> value_ids = new ArrayList<Integer>();
	
	private boolean pre_opengl20 = false;
	private boolean geom_shader = true;
	private boolean tess_shader = true;
	
	/**
	 * Constructor with initialization.
	 * 
	 * @param path
	 */
	public Shader(String path) {
		
		name = path;
		
		String vert_path = path + ".vert";
		String geom_path = path + ".geom";
		String frag_path = path + ".frag";
		String tess_control_path = path + ".tc";
		String tess_evaluation_path = path + ".te";
		
		//CHECK version
		String version = GL11.glGetString(GL11.GL_VERSION);
		if (Integer.valueOf(version.substring(0,1)) < 2) {
			//System.out.println("version too low");
			pre_opengl20 = true;
			if (!GLContext.getCapabilities().GL_ARB_shader_objects) 
				System.out.println("Pre OpenGL 2.0 without GL_ARB_shader_objects support.");
			if (!GLContext.getCapabilities().GL_ARB_vertex_shader) 
				System.out.println("Pre OpenGL 2.0 without GL_ARB_vertex_shader support.");
			if (!GLContext.getCapabilities().GL_ARB_fragment_shader) 
				System.out.println("Pre OpenGL 2.0 without GL_ARB_vertex_shader support.");
		}
		
		String line = "";
		BufferedReader reader;
		
		/** VERTEX SHADER */
		String vertex_shader_source = "";
		try {
			reader = new BufferedReader(FileIO.resolveAbsolutePathFR(vert_path));
			while( (line=reader.readLine()) != null){
				vertex_shader_source += line + "\n";
			}
			reader.close();
		} catch (IOException e1) {
			e1.printStackTrace();
			//throw new NullPointerException("Vertex Shader path not found.");
		}
		
		if (!pre_opengl20) {
			vertex_shader = compile(GL20.GL_VERTEX_SHADER, vertex_shader_source);
		} else {
			vertex_shader = compilePre20(ARBVertexShader.GL_VERTEX_SHADER_ARB, vertex_shader_source);
		}
		if (vertex_shader == 0) System.out.println("Error Vertex Shader!");
		
		
		/** GEOMETRY SHADER */
		String geometry_shader_source = "";
		try {
			InputStreamReader fr = FileIO.resolveAbsolutePathFR(geom_path);
			if (fr != null) {
				reader = new BufferedReader(fr);
				while( (line=reader.readLine()) != null){
					geometry_shader_source += line + "\n";
				}
				reader.close();
			} else {
				geom_shader = false;
			}
		} catch (IOException e1) {
			e1.printStackTrace();
		}
		//TODO: version checking...
		if (geom_shader) {
			geometry_shader = compile(GL32.GL_GEOMETRY_SHADER, geometry_shader_source);
			if (geometry_shader == 0) System.out.println("Error Geometry Shader!");
		}
		
		
		/** TESSELLATION CONTROL SHADER */
		String tess_control_shader_source = "";
		try {
			InputStreamReader fr = FileIO.resolveAbsolutePathFR(tess_control_path);
			if (fr != null) {
				reader = new BufferedReader(fr);
				while( (line=reader.readLine()) != null){
					tess_control_shader_source += line + "\n";
				}
				reader.close();
			} else {
				tess_shader = false;
			}
		} catch (IOException e1) {
			e1.printStackTrace();
		}
		//TODO: version checking...
		if (tess_shader) {
			tess_control_shader = compile(GL40.GL_TESS_CONTROL_SHADER, tess_control_shader_source);
			if (tess_control_shader == 0) System.out.println("Error Tessellation Control Shader!");
		}
		
		
		/** TESSELLATION EVALUATION SHADER */
		String tess_evaluation_shader_source = "";
		try {
			InputStreamReader fr = FileIO.resolveAbsolutePathFR(tess_evaluation_path);
			if (fr != null) {
				reader = new BufferedReader(fr);
				while( (line=reader.readLine()) != null){
					tess_evaluation_shader_source += line + "\n";
				}
				reader.close();
			} else {
				tess_shader = false;
			}
		} catch (IOException e1) {
			e1.printStackTrace();
		}
		//TODO: version checking...
		if (tess_shader) {
			tess_evaluation_shader = compile(GL40.GL_TESS_EVALUATION_SHADER, tess_evaluation_shader_source);
			if (tess_evaluation_shader == 0) System.out.println("Error Tessellation Evaluation Shader!");
		}
		
		if (debug) System.out.println("tessellation: " + tess_shader);
		
		/** FRAGMENT SHADER */
		String fragment_shader_source = "";
		try {
			reader = new BufferedReader(FileIO.resolveAbsolutePathFR(frag_path));
			while( (line=reader.readLine()) != null){
				fragment_shader_source += line + "\n";
			}
			reader.close();
		} catch (IOException e1) {
			e1.printStackTrace();
			//throw new NullPointerException("Fragment Shader path not found.");
		}
		
		if (!pre_opengl20) {
			fragment_shader = compile(GL20.GL_FRAGMENT_SHADER, fragment_shader_source);
		} else {
			fragment_shader = compilePre20(ARBFragmentShader.GL_FRAGMENT_SHADER_ARB, fragment_shader_source);
		}
		if (fragment_shader == 0) System.out.println("Error Fragment Shader!");
		
		
		/** LINK */
		
		if (!pre_opengl20) {
			link(path);
		} else {
			linkPre20(path);
		}
		
	}
	
	/**
	 * Compiles the shader
	 * 
	 * @param type
	 * @param shader_source
	 * @return
	 */
	private int compilePre20(int type, String shader_source) {
		
		int shader = ARBShaderObjects.glCreateShaderObjectARB(type);
		if (shader == 0) {
			System.out.println("Impossible to create shader object!");
			return 0;
		}
		
		ARBShaderObjects.glShaderSourceARB(shader, shader_source);
		ARBShaderObjects.glCompileShaderARB(shader);
		
		return shader;
	}
	
	/**
	 * Compiles the shader.
	 * 
	 * @param type
	 * @param shader_source
	 * @return
	 */
	private int compile(int type, String shader_source) {
		
		int shader = GL20.glCreateShader(type);
		if (shader == 0) {
			System.out.println("Impossible to create shader object!");
			return 0;
		}
		
		GL20.glShaderSource(shader, shader_source);
		GL20.glCompileShader(shader);
		
		return shader;
	}
	
	private void linkPre20(String source) {
		
		program = ARBShaderObjects.glCreateProgramObjectARB();
		
		ARBShaderObjects.glAttachObjectARB(program, vertex_shader);
		ARBShaderObjects.glAttachObjectARB(program, fragment_shader);
		
		ARBShaderObjects.glLinkProgramARB(program);
		
		IntBuffer result = BufferUtils.createIntBuffer(1);
		ARBShaderObjects.glGetObjectParameterARB(program, ARBShaderObjects.GL_OBJECT_LINK_STATUS_ARB, result);
		
		//System.out.println(result.get(0));
		
		if(result.get(0) != GL11.GL_TRUE){
			
			System.out.println("Shader compile error! \n\n" + 
					"in: " + source + "\n\n" +
					GL20.glGetProgramInfoLog(program, 5000));
		}
		
		String debugOut = GL20.glGetProgramInfoLog(program, 5000);
		if (debugOut.length() > 0) System.out.println(debugOut);
		
	}
	
	/**
	 * Links the program.
	 * 
	 * @param source
	 */
	private void link(String source) {
		
		program = GL20.glCreateProgram();
		
		GL20.glAttachShader(program, vertex_shader);
		if (geom_shader) GL20.glAttachShader(program, geometry_shader);
		if (tess_shader) {
			GL20.glAttachShader(program, tess_control_shader);
			GL20.glAttachShader(program, tess_evaluation_shader);
		}
		GL20.glAttachShader(program, fragment_shader);
		
		//TODO - find a more general solution... - temporary hack
		GL30.glBindFragDataLocation(program, 0, "out_Color");
		GL30.glBindFragDataLocation(program, 1, "out_ColorTwo");
		//GL30.glBindFragDataLocation(program, 0, "out_Color");
		//GL30.glBindFragDataLocation(program, 1, "out_ColorTwo");
		
		GL20.glLinkProgram(program);
		
		IntBuffer result = BufferUtils.createIntBuffer(1);
		GL20.glGetProgram(program, GL20.GL_LINK_STATUS, result);
		
		//System.out.println(result.get(0));
		
		if(result.get(0) != GL11.GL_TRUE){
			System.out.println("Shader compile error! \n\n" + 
					"in: " + source + "\n\n" +
					GL20.glGetProgramInfoLog(program, 5000));
		}
		
		String debugOut = GL20.glGetProgramInfoLog(program, 5000);
		if (debugOut.length() > 0) System.out.println(debugOut);
		
	}
	
	/**
	 * Binds a shader.
	 */
	public void bind() {
		if (!pre_opengl20) {
			GL20.glUseProgram(program);
			
			try {
				Util.checkGLError();
			} catch (org.lwjgl.opengl.OpenGLException e) {
				
				//System.out.println(e);
				
				//there is/can be some error before this call... (invalid enum)
			}
			
			//GL30.glBindFragDataLocation(program, 0, "out_Color");
			//GL30.glBindFragDataLocation(program, 1, "out_ColorTwo");
			
			//GL30.glBindFragDataLocation(program, 0, "out_Color");
			//GL30.glBindFragDataLocation(program, 1, "out_Normal");
			
			Util.checkGLError();
			
		} else {
			ARBShaderObjects.glUseProgramObjectARB(program);
		}
	}
	
	/**
	 * Unbinds a shader.
	 */
	public void unbind() {
		if (!pre_opengl20) {
			GL20.glUseProgram(0);
		} else {
			ARBShaderObjects.glUseProgramObjectARB(0);
		}
	}
	
	/**
	 * Gets a uniform location.
	 * 
	 * @param name
	 * @return
	 */
	public int initUniform(String name) {
		int location = 0;
		if (!pre_opengl20) {
			location = GL20.glGetUniformLocation(program, name);
		} else {
			location = ARBShaderObjects.glGetUniformLocationARB(program, name);
		}
		
		//if (!disable_error_msg && location == -1) System.out.println("Error binding shader int. (shown before added)");
		
		value_ids.add(location);
		
		if (debug) System.out.println("Added: " + name + " at " + (value_ids.size()-1) + " location: " + location);
		
		return value_ids.size()-1;
	}
	
	/**
	 * Gets an in/out location.
	 * 
	 * @param name
	 * @return
	 */
	public int initInOut(String name) {
		int location = 0;
		if (!pre_opengl20) {
			location = GL20.glGetAttribLocation(program, name);
		} else {
			//does this work?
			location = ARBVertexShader.glGetAttribLocationARB(program, name);
		}
		
		//if (!disable_error_msg && location == -1) System.out.println("Error binding shader int.");
		
		//System.out.println("Shader - initInOut: " + location + " at " + name);
		
		value_ids.add(location);
		
		if (debug) System.out.println("Added: " + name + " at " + (value_ids.size()-1) + " location: " + location);
		
		return value_ids.size()-1;
	}
	
	public int getInOut(int position_in_location_array) {
		return value_ids.get(position_in_location_array);
	}
	
	// --------------------- UNIFORMS ---------------------
	
	public void setValue1i(int position_in_location_array, int value) {
		
		if (!pre_opengl20) {
			GL20.glUniform1i(value_ids.get(position_in_location_array), value);
		} else {
			ARBShaderObjects.glUniform1iARB(value_ids.get(position_in_location_array), value);
		}
		
	}
	
	public void setValue1f(int position_in_location_array, float value) {
		
		if (!pre_opengl20) {
			GL20.glUniform1f(value_ids.get(position_in_location_array), value);
		} else {
			ARBShaderObjects.glUniform1fARB(value_ids.get(position_in_location_array), value);
		}
		
	}
	
	public void setValueVec2f(int position_in_location_array, float x, float y) {
		
		if (!pre_opengl20) {
			GL20.glUniform2f(value_ids.get(position_in_location_array), x, y);
		} else {
			ARBShaderObjects.glUniform2fARB(value_ids.get(position_in_location_array), x, y);
		}
		
	}
	
	public void setValueVec3f(int position_in_location_array, float x, float y, float z) {
		
		if (!pre_opengl20) {
			GL20.glUniform3f(value_ids.get(position_in_location_array), x, y, z);
		} else {
			ARBShaderObjects.glUniform3fARB(value_ids.get(position_in_location_array), x, y, z);
		}
		
	}
	
	public void setValueVec4f(int position_in_location_array, float x, float y, float z, float w) {
		
		if (!pre_opengl20) {
			GL20.glUniform4f(value_ids.get(position_in_location_array), x, y, z, w);
		} else {
			ARBShaderObjects.glUniform4fARB(value_ids.get(position_in_location_array), x, y, z, w);
		}
		
	}
	
	public void setValueMat3f(int position_in_location_array, float [] value) {
		
		matrix.clear();
		for(int i=0; i<9; i++) {
			matrix.put(value[i]);
		}
		matrix.flip();
		
		if (!pre_opengl20) {
			GL20.glUniformMatrix3(value_ids.get(position_in_location_array), true, matrix);
		} else {
			ARBShaderObjects.glUniformMatrix3ARB(value_ids.get(position_in_location_array), true, matrix);
		}
		
	}
	
	public void setValueMat4x3f(int position_in_location_array, FloatBuffer value) {
		
		//matrix.clear();
		//for(int i=0; i<16; i++) {
		//	matrix.put(value[i]);
		//}
		//matrix.flip();
		
		if (!pre_opengl20) {
			GL21.glUniformMatrix4x3(value_ids.get(position_in_location_array), true, value);
		} else {
			System.out.println("Shader: 4x4 Matrix not implemented!");
			//TODO!
			//ARBShaderObjects.glUniformMatrix4ARB(value_ids.get(position_in_location_array), true, value);
		}
		
	}
	
	public void setValueMat4f(int position_in_location_array, FloatBuffer value) {
		
		if (!pre_opengl20) {
			GL20.glUniformMatrix4(value_ids.get(position_in_location_array), true, value);
		} else {
			ARBShaderObjects.glUniformMatrix4ARB(value_ids.get(position_in_location_array), true, value);
		}
		
	}
	
	public void setValueMat4f(int position_in_location_array, float[] value) {
		
		//System.out.println("matrix what...");
		
		matrix.clear();
		matrix.put(value);
		matrix.position(0);
		
		//System.out.println(matrix.get());
		
		if (!pre_opengl20) {
			GL20.glUniformMatrix4(value_ids.get(position_in_location_array), true, matrix);
		} else {
			ARBShaderObjects.glUniformMatrix4ARB(value_ids.get(position_in_location_array), true, matrix);
		}
		
	}
	
	public void setValueFloatArray(int position_in_location_array, float[] data) {
		
		FloatBuffer fData = BufferUtils.createFloatBuffer(data.length);
		fData.put(data);
		fData.flip();
		
		if (!pre_opengl20) {
			GL20.glUniform1(value_ids.get(position_in_location_array), fData);
		} else {
			ARBShaderObjects.glUniform1ARB(value_ids.get(position_in_location_array), fData);
		}
	}
	
	public String getName() {
		return name;
	}
	
}
