package renderer.shader;

import java.util.ArrayList;
import java.util.HashMap;


import org.lwjgl.opengl.GL11;

import com.infovis.callgraph.CallBackInterface;


/**
 * Top level container and definition for shader.
 * 
 * Implements functionality of multiple Shaders and Variables to Shaders.
 * 
 * @author DarkMatt
 *
 */

public class ShaderCollection {

	/*
	 * Merely to integrate the available shader into the engine.
	 * 
	 * Possibility to automate this process later on.
	 */
	
	//shader
	public static final int BSPLINE = 0;
	
	private ShaderWrapper activeShader = null;
	
	/**
	 * Wraps the Shader to allow arbitrary parameters.
	 * Parameters are lazy initialized.
	 * 
	 * @author DarkMatt
	 *
	 */
	private class ShaderWrapper extends Shader {
		
		/** Shader uniform locations */
		private HashMap<String, Integer> shaderUniforms;
		
		public ShaderWrapper(String path) {
			super(path);
		}
		
		/**
		 * "Lazy initialization".
		 * 
		 * Requires String/value pairs
		 * 
		 * @param params
		 */
		public void setShaderParameters(Object ... params) {
			
			// Error handling
			if (params.length%2 != 0) throw new RuntimeException("ShaderCollection: Invalid number of Shader parameters.");
			
			// Initialize if necessary
			if (shaderUniforms == null) {
				shaderUniforms = new HashMap<String, Integer>();
				
				for (int i = 0; i < params.length; i+=2) {
					String uniformName = String.valueOf(params[i]);
					Integer location = this.initUniform(uniformName);
					
					if (location == -1) {
						System.out.println("ShaderCollection: Uniform " + uniformName + " not found in Shader file." );
					}
					
					shaderUniforms.put(uniformName, location);
				}
			}
			
			// Sets the Uniforms
			for (int i = 0; i < params.length; i+=2) {
				
				Integer location = shaderUniforms.get(params[i]);
				
				if (location != null) {
				
					if (params[i+1] instanceof Integer) {
						//System.out.println("Integer " + params[i+1]);
						this.setValue1i(location, (Integer)params[i+1]);
						
					} else if (params[i+1] instanceof Float) {
						//System.out.println("Float " + params[i+1]);
						this.setValue1f(location, (Float)params[i+1]);
						
					} else if (params[i+1] instanceof int[]) {
						int[] arr = (int[])params[i+1];
						if (arr.length == 2) {
							this.setValueVec2f(location, (float)arr[0], (float)arr[1]);
						}
						
					} else if (params[i+1] instanceof float[]) {
						float[] arr = (float[])params[i+1];
						if (arr.length == 2) {
							this.setValueVec2f(location, arr[0], arr[1]);
						} else if (arr.length == 3) {
							this.setValueVec3f(location, arr[0], arr[1], arr[2]);
						} else if (arr.length == 16) {
							this.setValueMat4f(location, arr);
						} else {
							this.setValueFloatArray(location, arr);
						}
						
					} else {
						System.out.println("ShaderCollection: Type " + params[i+1].getClass().getName() + " not recognized.");
					}
					
				} else {
					
					System.out.println("ShaderColelction: Variable " + params[i] + " does not exsist in shader " + this.getName());
					
				}
				
			}
			
		}
		
	}
	
	private ArrayList<ShaderWrapper> shader;
	private int basic_in_Position, basic_in_Color;
	
	
	public ShaderCollection() {
		shader = new ArrayList<ShaderWrapper>();
		
		shader.add(new ShaderWrapper("icons/shader/bspline"));
    	basic_in_Position = shader.get(BSPLINE).initInOut("in_Position");
    	basic_in_Color = shader.get(BSPLINE).initInOut("in_Color");
    	
	}
	
	/**
	 * Using varargs.
	 * 
	 * Use: ("shaderNameOfVariable", value, ...)
	 * 
	 * @param params
	 */
	public void setShaderParameters(Object ... params) {
		
		activeShader.setShaderParameters(params);
		
	}
	
	public int getVertexLocation(int type) {
		switch (type) {
			case BSPLINE:
				//the index is not the index inside the shader!
				return shader.get(type).getInOut(basic_in_Position);
			default:
				return -1;
		}
	}
	
	public int getNormalLocation(int type) {
		switch (type) {
			default:
				return -1;
		}
	}
	
	public int getTextCoordLocation(int type) {
		switch (type) {
			default:
				return -1;
		}
	}
	
	public int getColorLocation(int type) {
		switch (type) {
			case BSPLINE:
				return shader.get(type).getInOut(basic_in_Color);
			default:
				return -1;
		}
	}
	
	public int getmodelMatrixLocation(int type) {
		switch (type) {
			default:
				return -1;
		}
	}
	
	/**
	 * Binds a specific shader.
	 * 
	 * @param name
	 */
	public void bind(int name) {
		//GL11.glPolygonMode(GL11.GL_FRONT_AND_BACK, GL11.GL_LINE);
		GL11.glPolygonMode(GL11.GL_FRONT_AND_BACK, GL11.GL_FILL);
		shader.get(name).bind();
		activeShader = shader.get(name);
	}
	
	/**
	 * Unbinds a specific shader.
	 * 
	 * @param name
	 */
	public void unbind(int name) {
		shader.get(name).unbind();
	}
	
	/**
	 * Unbinds all shaders.
	 */
	public void unbindAll() {
		for (int i=0; i<shader.size(); i++)
			shader.get(i).unbind();
	}
	
}
